"""Define project metadata
"""

__title__ = 'pydash'
__summary__ = ('A utility library for doing "stuff" in a functional way. '
               'Based on the Lo-Dash Javascript library.')
__url__ = 'https://github.com/dgilland/pydash'

__version__ = '2.3.0'

__install_requires__ = []

__author__ = 'Derrick Gilland'
__email__ = 'dgilland@gmail.com'

__license__ = 'MIT License'
